/*!\file StochasticForcingx
 * \brief: compute noise terms for the StochasticForcing fields
 */

#include "./StochasticForcingx.h"
#include "../../classes/Loads/Friction.h"
#include "../../shared/shared.h"
#include "../../toolkits/toolkits.h"
#include "../../shared/Random/random.h"

void StochasticForcingx(FemModel* femmodel){/*{{{*/


	//VV testing (12Nov)
	/*
	IssmDouble timeVV,dtVV,starttimeVV;
	femmodel->parameters->FindParam(&timeVV,TimeEnum);
   femmodel->parameters->FindParam(&dtVV,TimesteppingTimeStepEnum);
   femmodel->parameters->FindParam(&starttimeVV,TimesteppingStartTimeEnum);
	IssmDouble valMean = 0;
	IssmDouble valSdev = 0.5;
	int seed;
	//seed = reCast<int,IssmDouble>((timeVV-starttimeVV)/dtVV);
	seed = -1;
	IssmDouble rdmVV;
	univariateNormal_test0(&rdmVV,valMean,valSdev,seed);
	_printf_("VV rdmVV: "<<rdmVV<<'\n');
	*/

   /*Retrieve parameters*/
   bool randomflag;
   int M,N,numfields,my_rank;
   int* fields            = NULL;
   int* dimensions        = NULL;
   IssmDouble* covariance = NULL;
   femmodel->parameters->FindParam(&randomflag,StochasticForcingRandomflagEnum);
   femmodel->parameters->FindParam(&numfields,StochasticForcingNumFieldsEnum);
   femmodel->parameters->FindParam(&fields,&N,StochasticForcingFieldsEnum);    _assert_(N==numfields);
   femmodel->parameters->FindParam(&dimensions,&N,StochasticForcingDimensionsEnum);    _assert_(N==numfields);
   int dimtot=0;
   for(int i=0;i<numfields;i++) dimtot = dimtot+dimensions[i];
   femmodel->parameters->FindParam(&covariance,&M,&N,StochasticForcingCovarianceEnum); _assert_(M==dimtot); _assert_(N==dimtot);

   /*Compute noise terms*/
   IssmDouble* noiseterms = xNew<IssmDouble>(dimtot);
   my_rank=IssmComm::GetRank();
   if(my_rank==0){
      int fixedseed;
      IssmDouble time,dt,starttime;
      femmodel->parameters->FindParam(&time,TimeEnum);
      femmodel->parameters->FindParam(&dt,TimesteppingTimeStepEnum);
      femmodel->parameters->FindParam(&starttime,TimesteppingStartTimeEnum);
		/*Determine whether random seed is fixed to time step (randomflag==false) or random seed truly random (randomflag==true)*/
      if(randomflag) fixedseed=-1;
      else fixedseed = reCast<int,IssmDouble>((time-starttime)/dt);
		/*multivariateNormal needs to be passed a NULL pointer to avoid memory leak issues*/
      IssmDouble* temparray = NULL;
      multivariateNormal(&temparray,dimtot,0.0,covariance,fixedseed);
      for(int i=0;i<dimtot;i++) noiseterms[i]=temparray[i];
		xDelete<IssmDouble>(temparray);
   }
   ISSM_MPI_Bcast(noiseterms,dimtot,ISSM_MPI_DOUBLE,0,IssmComm::GetComm());
   
	int i=0;
   for(int j=0;j<numfields;j++){
      int dimenum_type,noiseenum_type;
      IssmDouble* noisefield = xNew<IssmDouble>(dimensions[j]);
      for(int k=0;k<dimensions[j];k++){
         noisefield[k]=noiseterms[i+k];
      }
     
		int dimensionid;

		/*Deal with the autoregressive models*/
		if(fields[j]==SMBautoregressionEnum || fields[j]==FrontalForcingsRignotAutoregressionEnum){
			switch(fields[j]){
				case SMBautoregressionEnum:
					dimenum_type   = SmbBasinsIdEnum;
					noiseenum_type = SmbAutoregressionNoiseEnum;
					break;
				case FrontalForcingsRignotAutoregressionEnum:
					dimenum_type   = FrontalForcingsBasinIdEnum;
					noiseenum_type = ThermalforcingAutoregressionNoiseEnum;
					break;
			}
			for(Object* &object:femmodel->elements->objects){
            Element* element = xDynamicCast<Element*>(object);
            int numvertices  = element->GetNumberOfVertices();
            IssmDouble* noise_element = xNew<IssmDouble>(numvertices);
            element->GetInputValue(&dimensionid,dimenum_type);
            for(int i=0;i<numvertices;i++) noise_element[i] = noisefield[dimensionid];
            element->AddInput(noiseenum_type,noise_element,P0Enum);
            xDelete<IssmDouble>(noise_element);
			}
		}
		else{
			switch(fields[j]){
				case SMBautoregressionEnum:
				case FrontalForcingsRignotAutoregressionEnum:
					/*Already done above*/
					break;
				case DefaultCalvingEnum:
					/*Delete CalvingCalvingrateEnum at previous time step (required if it is transient)*/
					femmodel->inputs->DeleteInput(CalvingCalvingrateEnum);
					for(Object* &object:femmodel->elements->objects){
						Element* element = xDynamicCast<Element*>(object);
						int numvertices  = element->GetNumberOfVertices();
						IssmDouble baselinecalvingrate;
						IssmDouble calvingrate_tot[numvertices];
						Input* baselinecalvingrate_input  = NULL;
						baselinecalvingrate_input = element->GetInput(BaselineCalvingCalvingrateEnum); _assert_(baselinecalvingrate_input);
						element->GetInputValue(&dimensionid,StochasticForcingDefaultIdEnum);
						Gauss* gauss = element->NewGauss();
						for(int i=0;i<numvertices;i++){
							gauss->GaussVertex(i);
							baselinecalvingrate_input->GetInputValue(&baselinecalvingrate,gauss);
							calvingrate_tot[i] = max(0.0,baselinecalvingrate+noisefield[dimensionid]);
						}
						element->AddInput(CalvingCalvingrateEnum,&calvingrate_tot[0],P1DGEnum);
						delete gauss;
					}
					break;
				case FloatingMeltRateEnum:
					/*Delete BasalforcingsFloatingiceMeltingRateEnum at previous time step (required if it is transient)*/
					femmodel->inputs->DeleteInput(BasalforcingsFloatingiceMeltingRateEnum);
					for(Object* &object:femmodel->elements->objects){
						Element* element = xDynamicCast<Element*>(object);
						int numvertices  = element->GetNumberOfVertices();
						IssmDouble baselinefloatingicemeltrate;
						IssmDouble floatingicemeltrate_tot[numvertices];
						Input* baselinefloatingicemeltrate_input  = NULL;
						baselinefloatingicemeltrate_input = element->GetInput(BaselineBasalforcingsFloatingiceMeltingRateEnum); _assert_(baselinefloatingicemeltrate_input);
						element->GetInputValue(&dimensionid,StochasticForcingDefaultIdEnum);
						Gauss* gauss = element->NewGauss();
						for(int i=0;i<numvertices;i++){
							gauss->GaussVertex(i);
							baselinefloatingicemeltrate_input->GetInputValue(&baselinefloatingicemeltrate,gauss);
							/*No check for positive melt rate because basal accretion is allowed*/
							floatingicemeltrate_tot[i] = baselinefloatingicemeltrate+noisefield[dimensionid];
						}
						element->AddInput(BasalforcingsFloatingiceMeltingRateEnum,&floatingicemeltrate_tot[0],P1DGEnum);
						delete gauss;
					}
					break;
				case SMBforcingEnum:
					/*Delete SmbMassBalanceEnum at previous time step (required if it is transient)*/
					femmodel->inputs->DeleteInput(SmbMassBalanceEnum);
					for(Object* &object:femmodel->elements->objects){
						Element* element = xDynamicCast<Element*>(object);
						int numvertices  = element->GetNumberOfVertices();
						IssmDouble baselinesmb;
						IssmDouble smb_tot[numvertices];
						Input* baselinesmb_input  = NULL;
						baselinesmb_input = element->GetInput(BaselineSmbMassBalanceEnum); _assert_(baselinesmb_input);
						element->GetInputValue(&dimensionid,StochasticForcingDefaultIdEnum);
						Gauss* gauss = element->NewGauss();
						for(int i=0;i<numvertices;i++){
							gauss->GaussVertex(i);
							baselinesmb_input->GetInputValue(&baselinesmb,gauss);
							smb_tot[i] = baselinesmb+noisefield[dimensionid];
						}
						element->AddInput(SmbMassBalanceEnum,&smb_tot[0],P1DGEnum);
						delete gauss;
					}
					break;
				//VV working(29Nov2021)
				case FrictionWaterPressureEnum:
					/*Specify that WaterPressure is stochastic*/ 
					femmodel->parameters->SetParam(true,StochasticForcingIsWaterPressureEnum);
					for(Object* &object:femmodel->elements->objects){
                  Element* element = xDynamicCast<Element*>(object);
                  int numvertices  = element->GetNumberOfVertices();
                  IssmDouble p_water_deterministic[numvertices];
                  IssmDouble p_water[numvertices];
						element->GetInputValue(&dimensionid,StochasticForcingDefaultIdEnum);
						Gauss* gauss=element->NewGauss();
						Friction* friction = new Friction(element);
						for(int i=0;i<numvertices;i++){
							gauss->GaussVertex(i);
							p_water_deterministic[i] = friction->SubglacialWaterPressure(gauss);
							p_water[i]               = p_water_deterministic[i] + noisefield[dimensionid]; //make sure positive (29Nov2021)
							p_water[i]               = max(0.0,p_water[i]);
						}
						element->AddInput(FrictionWaterPressureEnum,p_water,P1DGEnum);
						delete gauss;
						delete friction;
					}
					break;

				/*
				case FrictionEffectivePressureEnum:
					femmodel->parameters->SetParam(true,StochasticForcingIsEffectivePressureEnum);
					for(Object* &object:femmodel->elements->objects){
                  Element* element = xDynamicCast<Element*>(object);
                  int numvertices  = element->GetNumberOfVertices();
                  IssmDouble Neff[numvertices];
						element->GetInputValue(&dimensionid,StochasticForcingDefaultIdEnum);
						Gauss* gauss=element->NewGauss();
						Friction* friction = new Friction(element);
						for(int i=0;i<numvertices;i++){
							gauss->GaussVertex(i);
							Neff[i] = friction->EffectivePressure(gauss);
							Neff[i] = Neff[i]+noisefield[dimensionid];
						}
						element->AddInput(FrictionEffectivePressureEnum,Neff,P1DGEnum);
						delete gauss;
						delete friction;
					}
					break;
				*/

				default:
					_error_("Field "<<EnumToStringx(fields[j])<<" does not support stochasticity yet.");
			}
		}
		i=i+dimensions[j];
      xDelete<IssmDouble>(noisefield);
   }

	/*Cleanup*/
   xDelete<int>(fields);
   xDelete<int>(dimensions);
   xDelete<IssmDouble>(covariance);
   xDelete<IssmDouble>(noiseterms);
}/*}}}*/
